package fire;

// Fire model for SFI Graduate Workshop in Computational Economics
// (c)2003 Laszlo Gulyas, Scott Moser, Jason Woodard

import uchicago.src.sim.gui.*;
import java.awt.*;
import java.util.Iterator;
import java.util.Vector;

public class Agent implements Drawable {

    // Agent parameters
    int ID;
    int emergencyVision, moveVision;

    // Agent variables
    boolean isEmergency;

    // Position
    int x, y;
    int newX, newY;
    int oldX, oldY;

    // Perception
    static final int FIRE_DIR = 0;
    static final int DOOR_DIR = 1;
    static final int WALL_DIR = 2;
    static final int AVG_DIR = 3;
    static final int RND_DIR = 4;
    DirVector[] perception;

    // Rule(s)
    MovingRule rule;

    // Technical parameters
    Model model;

    public Agent(int ID, int x, int y, int emergencyVision, int moveVision, Model model) {
        this.ID = ID;
        this.x = x;
        this.y = y;
        this.emergencyVision = emergencyVision;
        this.moveVision = moveVision;
        this.model = model;

        isEmergency = false;

        perception = new DirVector[5];

        rule = new SimpleRule(5);

        oldX = x;
        oldY = y;
    }

    ////////////////////////////////////////////////////////////////////////////////

    protected void checkIfEmergency() {
        Vector neighbors = model.world.getMooreNeighbors(x, y,
                emergencyVision,
                emergencyVision,
                false
        );

        Iterator it = neighbors.iterator();
        while ((it.hasNext()) && !isEmergency) {
            Object o = it.next();
            if (o instanceof Fire) {
                isEmergency = true;
            } else if (o instanceof Agent) {
                Agent anAgent = (Agent)o;
                isEmergency = anAgent.isEmergency;
            }
        }
    }

    ////////////////////////////////////////////////////////////////////////////////

    int[] diffX = {0, 0, 1, 1, 1, 0, -1, -1, -1};
    int[] diffY = {0, -1, -1, 0, 1, 1, 1, 0, -1};

    protected void determineNormalMove() {
        int pos = model.getNextIntFromTo(0, 8);
        newX = x + diffX[pos];
        newY = y + diffY[pos];
    }

    ////////////////////////////////////////////////////////////////////////////////

    protected void getPerception() {
        Vector neighbors = model.world.getMooreNeighbors(x, y,
                moveVision, moveVision,
                false);

        // Add ourselves to the list to make sure we count ourselves in the avg
        neighbors.add(this);

        DirVector minFire = DirVector.NULL;
        boolean foundFire = false;
        DirVector minDoor = DirVector.NULL;
        boolean foundDoor = false;
        DirVector minWall = DirVector.NULL;
        boolean foundWall = false;
        DirVector avg = DirVector.NULL;

        Iterator it = neighbors.iterator();
        while (it.hasNext()) {
            Object o = it.next();
            if (o instanceof Fire) {
                Fire f = (Fire)o;
                DirVector v = new DirVector(new Point(x, y), new Point(f.x, f.y));
                if ((v.getLength() < minFire.getLength()) || !foundFire) {
                    minFire = v;
                    foundFire = true;
                }
            } else if (o instanceof Door) {
                Door d = (Door)o;
                DirVector v = new DirVector(new Point(x, y), new Point(d.x, d.y));
                if ((v.getLength() < minDoor.getLength()) || !foundDoor) {
                    minDoor = v;
                    foundDoor = true;
                }
            } else if (o instanceof Agent) {
                Agent a = (Agent)o;
                DirVector v = new DirVector(new Point(a.oldX, a.oldY), new Point(a.x, a.y));
                avg = avg.add(v);
            } else if (o instanceof Wall) {
                Wall w = (Wall)o;
                DirVector v = new DirVector(new Point(x, y), new Point(w.x, w.y));
                if ((v.getLength() < minWall.getLength()) || !foundWall) {
                    minWall = v;
                    foundWall = true;
                }
            }
        }

        perception[FIRE_DIR] = minFire;
        perception[DOOR_DIR] = minDoor;
        perception[WALL_DIR] = minWall;
        perception[AVG_DIR] = avg;
        perception[RND_DIR] = new DirVector(model.getNextDoubleFromTo(0, 360));
    }

    protected void determineEscapeMove() {
        getPerception();
        DirVector v = rule.getDirection(perception);
        Point p = v.getMoveFrom(new Point(x, y));
        newX = p.x;
        newY = p.y;
    }

    ////////////////////////////////////////////////////////////////////////////////

    public void decide() {
        checkIfEmergency();

        if (isEmergency) {
            determineEscapeMove();
        } else {
            determineNormalMove();
        }
    }

    ////////////////////////////////////////////////////////////////////////////////
    // Methods required by the Drawable interface
    ////////////////////////////////////////////////////////////////////////////////

    public int getX() {
        return x;
    }

    public int getY() {
        return y;
    }

    public void draw(SimGraphics g) {
        if (isEmergency) {
            g.drawFastRoundRect(Color.blue);
        } else {
            g.drawFastRoundRect(Color.white);
        }
    }
}