package fire;

// Fire model for SFI Graduate Workshop in Computational Economics
// (c)2003 Laszlo Gulyas, Scott Moser, Jason Woodard

import uchicago.src.sim.engine.*;
import uchicago.src.sim.space.*;
import uchicago.src.sim.util.*;
import java.awt.*;
import java.util.*;

public class Model extends SimpleModel {

    // Model parameters
    int numAgents;
    int emergencyVision, minMoveVision, maxMoveVision;
    int worldSize;

    int[][] doors;
    int[][] fire;
    int[][] walls;
    ArrayList fireList;
    int fireGrowthRate;

    // Model variables
    Object2DGrid world;
    int numDead, numEscaped;

    // Constants for locating empty neighbors
    int[] diffX = new int[]{-1, 0, 1, -1, 1, -1, 0, 1};
    int[] diffY = new int[]{-1, -1, -1, 0, 0, 1, 1, 1};

    public Model() {
        numAgents = 100; //20;
        worldSize = 50;
        emergencyVision = 50;
        minMoveVision = 10;
        maxMoveVision = 10;

        fireGrowthRate = 10;

        doors = new int[][]{{1, 0}, {2, 0}, {3, 0}, {4, 0}};
        fire = new int[][]{{24, 24}, {24, 25}, {25, 24}, {25, 25}};

        walls = new int[4 * worldSize - 4][2];
        int pos = 0;
        for (int i = 0; i < worldSize; i++) {
            walls[pos][0] = i;
            walls[pos][1] = 0;
            pos++;
            walls[pos][0] = i;
            walls[pos][1] = worldSize - 1;
            pos++;
        }
        for (int i = 1; i < worldSize - 1; i++) {
            walls[pos][0] = 0;
            walls[pos][1] = i;
            pos++;
            walls[pos][0] = worldSize - 1;
            walls[pos][1] = i;
            pos++;
        }
    }

    public void setup() {
        super.setup();
    }

    public void buildModel() {
        super.buildModel();

        world = new Object2DGrid(worldSize, worldSize);

        // Putting in the walls
        for (int i = 0; i < walls.length; i++) {
            int x = walls[i][0];
            int y = walls[i][1];
            world.putObjectAt(x, y, new Wall(x, y));
        }

        // Putting in the doors
        for (int i = 0; i < doors.length; i++) {
            int x = doors[i][0];
            int y = doors[i][1];
            world.putObjectAt(x, y, new Door(x, y));
        }

        // Putting in the fire(s)
        fireList = new ArrayList();
        for (int i = 0; i < fire.length; i++) {
            int x = fire[i][0];
            int y = fire[i][1];
            Fire f = new Fire(x, y);
            world.putObjectAt(x, y, f);
            fireList.add(f);
        }

        // Create the agents
        for (int i = 0; i < numAgents; i++) {
            int x, y;
            do {
                x = this.getNextIntFromTo(0, worldSize - 1);
                y = this.getNextIntFromTo(0, worldSize - 1);
            } while (world.getObjectAt(x, y) != null);

            int vis = this.getNextIntFromTo(minMoveVision, maxMoveVision);
            Agent anAgent = new Agent(i, x, y, emergencyVision, vis, this);
            agentList.add(anAgent);
            world.putObjectAt(x, y, anAgent);
        }

        numDead = 0;
        numEscaped = 0;
    }

    protected void moveIt(Agent anAgent) {
        world.putObjectAt(anAgent.x, anAgent.y, null);
        world.putObjectAt(anAgent.newX, anAgent.newY, anAgent);
        anAgent.oldX = anAgent.x;
        anAgent.oldY = anAgent.y;
        anAgent.x = anAgent.newX;
        anAgent.y = anAgent.newY;
    }

    protected void executeMoves() {
        Iterator it = agentList.iterator();
        while (it.hasNext()) {
            Agent anAgent = (Agent)it.next();

            if ((anAgent.newX > -1) && (anAgent.newX < worldSize) &&
                    (anAgent.newY > -1) && (anAgent.newY < worldSize)) {
                Object o = world.getObjectAt(anAgent.newX, anAgent.newY);

                if (o == null) {
                    // empty location ==> move there
                    moveIt(anAgent);
                } else if (o instanceof Fire) {
                    // the agent dies
                    it.remove();
                    numDead++;
                    world.putObjectAt(anAgent.x, anAgent.y, null);
                } else if (o instanceof Door) {
                    // the agent escapes and lives happily
                    it.remove();
                    numEscaped++;
                    world.putObjectAt(anAgent.x, anAgent.y, null);
                } else {
                    // try to move randomly

                    // This will only work if there is a wall around the boundary,
                    // as the naive mapping we do with diffX/Y won't work if you
                    // don't get all 8 neighbors...
                    Vector neighbors = world.getMooreNeighbors(anAgent.x, anAgent.y, true);
                    ArrayList list = new ArrayList();

                    int pos = 0;
                    Iterator itt = neighbors.iterator();
                    while (itt.hasNext()) {
                        Object oo = itt.next();
                        if (oo == null) {
                            list.add(new Point(anAgent.x + diffX[pos], anAgent.y + diffY[pos]));
                        }
                        pos++;
                    }

                    if (list.size() > 0) {
                        int rnd = this.getNextIntFromTo(0, list.size() - 1);
                        Point p = (Point)list.get(rnd);
                        anAgent.newX = p.x;
                        anAgent.newY = p.y;
                        moveIt(anAgent);
                    }
                }
            }
        }
    }

    protected void growFire() {
        ArrayList growth = new ArrayList();

        Iterator i = fireList.iterator();
        while (i.hasNext()) {
            Fire f = (Fire)i.next();
            for (int x = f.x - 1; x < f.x + 2; x++) {
                for (int y = f.y - 1; y < f.y + 2; y++) {
                    if ((x > -1) && (x < worldSize) && (y > -1) && (y < worldSize)) {
                        Object o = world.getObjectAt(x, y);
                        if ((o == null) || (o instanceof Agent)) {
                            if (o != null) {
                                numDead++;
                                agentList.remove(o);
                            }
                            Fire ff = new Fire(x, y);
                            world.putObjectAt(x, y, ff);
                            growth.add(ff);
                        }
                    }
                }
            }
        }

        fireList.addAll(growth);
    }

    public void step() {

        // Fire spread
        if (this.getTickCount() % fireGrowthRate == 1) {
            growFire();
        }

        // Randomize the order of execution
        SimUtilities.shuffle(agentList);

        // Let all the agents decide on where to move
        Iterator it = agentList.iterator();
        while (it.hasNext()) {
            Agent anAgent = (Agent)it.next();
            anAgent.decide();
        }

        // Let's execute the moves
        executeMoves();
    }

    ////////////////////////////////////////////////////////////////////////////////

    public static void main(String[] args) {
        SimInit init = new SimInit();
        init.loadModel(new Model(), "", false);
    }
}